@@ -12,11 +12,18 @@ use strict;
 
 sub get_pdls {my($this) = @_; return ($this->{ParNames},$this->{ParObjs});}
 
+my @code_args_always = qw(SignatureObj GenericTypes ExtraGenericLoops HaveThreading Name);
+sub make_args {
+  my ($which) = @_;
+  ("Parsed$which", [$which,"_Bad$which",@code_args_always]);
+}
+
 # Do the appropriate substitutions in the code.
 sub new {
-    my($type,$code,$badcode,$parnames,$parobjs,$indobjs,$generictypes,
-       $extrageneric,$havethreading,$name,
+    my($class,$code,$badcode,
+       $sig,$generictypes,$extrageneric,$havethreading,$name,
        $dont_add_thrloop, $backcode ) = @_;
+    my $parnames = $sig->names_sorted;
 
     die "Error: missing name argument to PDL::PP::Code->new call!\n"
       unless defined $name;
@@ -31,7 +38,7 @@ sub new {
     # "backcode" is a flag to the PDL::PP::Threadloop class indicating thre threadloop
     #   is for writeback code (typically used for writeback of data from child to parent PDL
 
-    $dont_add_thrloop = 0 unless defined $dont_add_thrloop;
+    $dont_add_thrloop ||= !$havethreading; # two have identical (though inverted) meaning so only track one
 
     # C++ style comments
     #
@@ -45,36 +52,35 @@ sub new {
 	print "DONT_ADD_THRLOOP!\n" if $dont_add_thrloop;
 	print "EXTRAGEN: {" .
 	  join(" ",
-	       map { "$_=>" . $$extrageneric{$_}} keys %$extrageneric)
+	       map "$_=>$$extrageneric{$_}", keys %$extrageneric)
 	    . "}\n";
 	print "ParNAMES: ",(join ',',@$parnames),"\n";
 	print "GENTYPES: ", @$generictypes, "\n";
 	print "HandleBad: $handlebad\n";
     }
     my $this = bless {
-	IndObjs => $indobjs,
+	IndObjs => $sig->dims_obj,
 	ParNames => $parnames,
-	ParObjs => $parobjs,
+	ParObjs => $sig->objs,
+	Sig => $sig,
 	Gencurtype => [], # stack to hold GenType in generic loops
 	types => 0,  # for thisisloop method
 	ftypes_vars => {},
 	ftypes_type => undef,
         Generictypes => $generictypes,   # so that MacroAccess can check it
         Name => $name,
-    }, $type;
-
-    my $inccode = join '',map {$_->get_incregisters();} (sort values %{$this->{ParObjs}});
+    }, $class;
 
     # First, separate the code into an array of C fragments (strings),
     # variable references (strings starting with $) and
     # loops (array references, 1. item = variable.
     #
     my ( $threadloops, $coderef, $sizeprivs ) =
-	$this->separate_code( "{$inccode\n$code\n}" );
+	$this->separate_code( "{$code\n}" );
 
     # Now, if there is no explicit threadlooping in the code,
     # enclose everything into it.
-    if(!$threadloops && !$dont_add_thrloop && $havethreading) {
+    if(!$threadloops && !$dont_add_thrloop) {
 	print "Adding threadloop...\n" if $::PP_VERBOSE;
 	my $nc = $coderef;
 	if( !$backcode ){ # Normal readbackdata threadloop
@@ -95,9 +101,9 @@ sub new {
     if ( $handlebad ) {
 	print "Processing 'bad' code...\n" if $::PP_VERBOSE;
 	my ( $bad_threadloops, $bad_coderef, $bad_sizeprivs ) =
-	    $this->separate_code( "{$inccode\n$badcode\n}" );
+	    $this->separate_code( "{$badcode\n}" );
 
-	if(!$bad_threadloops && !$dont_add_thrloop && $havethreading) {
+	if(!$bad_threadloops && !$dont_add_thrloop) {
 	    print "Adding 'bad' threadloop...\n" if $::PP_VERBOSE;
 	    my $nc = $bad_coderef;
 	    if( !$backcode ){ # Normal readbackdata threadloop
@@ -149,15 +155,27 @@ sub new {
 	push @$coderef,$nc;
     }
 
+    my $pobjs = $sig->objs;
     # Then, in this form, put it together what we want the code to actually do.
     print "SIZEPRIVS: ",(join ',',%$sizeprivs),"\n" if $::PP_VERBOSE;
-    $this->{Code} = "{".(join '',sort values %$sizeprivs).
+    $this->{Code} = (join '',sort values %$sizeprivs).
+       ($dont_add_thrloop?'':PDL::PP::pp_line_numbers __LINE__, join "\n",
+        'PDL_COMMENT("threadloop declarations")',
+        'register PDL_Indx __tind0,__tind1; PDL_COMMENT("counters along dim")',
+        'register PDL_Indx __tnpdls = $PRIV(pdlthread).npdls;',
+        'PDL_COMMENT("dims here are how many steps along those dims")',
+        (map "register PDL_Indx __tinc0_$_ = \$PRIV(pdlthread).incs[$_];", 0..$#$parnames),
+        (map "register PDL_Indx __tinc1_$_ = \$PRIV(pdlthread).incs[__tnpdls+$_];", 0.. $#$parnames),
+       ).
+       join('',map $_->get_incregisters, @$pobjs{sort keys %$pobjs}).
        $coderef->get_str($this,[])
-       ."}";
+       ;
     $this->{Code};
 
 } # new()
 
+sub sig {$_[0]->{Sig}}
+
 # This sub determines the index name for this index.
 # For example, a(x,y) and x0 becomes [x,x0]
 sub make_loopind { my($this,$ind) = @_;
@@ -265,7 +283,7 @@ my $line_re = qr/#\s*line\s+(\d+)\s+"([^"]*)"/;
 sub report_error {
     my ($message, $code) = @_;
     # Just croak if they didn't supply a #line directive:
-    croak($message) if $code !~ $line_re;
+    confess($message) if $code !~ $line_re;
     # Find the line at which the error occurred:
     my $line = 0;
     my $filename;
@@ -336,7 +354,7 @@ sub get_str_int {
       my $it = $this->myitem($parent,$nth);
       last MYLOOP if $nth and !$it;
       $str .= $it;
-      $str .= (join '',map {ref $_ ? $_->get_str($parent,$context) : $_}
+      $str .= join('', map ref $_ ? $_->get_str($parent,$context) : $_,
 	       @{$this}[$this->myoffs()..$#{$this}]);
       $nth++;
   }
@@ -383,8 +401,9 @@ sub new { my($type,$args,$sizeprivs,$parent) = @_;
 	for(@{$this->[0]}) {
 		print "SIZP $sizeprivs, $_\n" if $::PP_VERBOSE;
 		my $i = $parent->make_loopind($_);
+		my $i_size = $parent->sig->ind_obj($i->[0])->get_size;
 		$sizeprivs->{$i->[0]} =
-		  "register PDL_Indx __$i->[0]_size = \$PRIV(__$i->[0]_size);\n";
+		  "register PDL_Indx __$i->[0]_size = $i_size;\n";
 		print "SP :",(join ',',%$sizeprivs),"\n" if $::PP_VERBOSE;
 	}
 	return $this;
@@ -392,9 +411,9 @@ sub new { my($type,$args,$sizeprivs,$parent) = @_;
 
 sub myoffs { return 1; }
 sub myprelude { my($this,$parent,$context) = @_;
-	my $text = ""; my $i;
+	my $text = "";
 	push @$context, map {
-		$i = $parent->make_loopind($_);
+		my $i = $parent->make_loopind($_);
 # Used to be $PRIV(.._size) but now we have it in a register.
 		$text .= "{PDL_COMMENT(\"Open $_\") register PDL_Indx $_;
 			for($_=0; $_<(__$i->[0]_size); $_++) {";
@@ -404,7 +423,7 @@ sub myprelude { my($this,$parent,$context) = @_;
 }
 sub mypostlude { my($this,$parent,$context) = @_;
 	splice @$context, - ($#{$this->[0]}+1);
-	return join '',map {"}} PDL_COMMENT(\"Close $_\")"} @{$this->[0]};
+	return join '', map "}} PDL_COMMENT(\"Close $_\")", @{$this->[0]};
 }
 
 ###########################
@@ -468,36 +487,6 @@ sub mypostlude {
 	default:barf(\"PP INTERNAL ERROR in $parent->{Name}: unhandled datatype(%d), only handles ($supported)! PLEASE MAKE A BUG REPORT\\n\", $this->[3]);}\n";
 }
 
-
-package PDL::PP::SimpleThreadLoop;
-use Carp;
-our @ISA = "PDL::PP::Block";
-our @CARP_NOT;
-
-sub new { my($type) = @_; bless [],$type; }
-sub myoffs { return 0; }
-sub myprelude {my($this,$parent,$context) = @_;
- my $no;
- my ($ord,$pdls) = $parent->get_pdls();
-PDL::PP::pp_line_numbers(__LINE__, '	PDL_COMMENT("THREADLOOPBEGIN")
- if(PDL->startthreadloop(&($PRIV(__pdlthread)),$PRIV(vtable)->readdata,
- 	__privtrans))) return;
-   do {
- '.(join '',map {"${_}_datap += \$PRIV(__pdlthread).offs[".(0+$no++)."];\n"}
- 		@$ord).'
-');
-}
-
-sub mypostlude {my($this,$parent,$context) = @_;
- my $no;
- my ($ord,$pdls) = $parent->get_pdls();
-'	PDL_COMMENT("THREADLOOPEND")
- '.(join '',map {"${_}_datap -= \$PRIV(__pdlthread).offs[".(0+$no++)."];\n"}
- 		@$ord).'
-      } while(PDL->iterthreadloop(&$PRIV(__pdlthread),0));
- '
-}
-
 ####
 #
 # This relies on PP.pm making sure that initthreadloop always sets
@@ -508,7 +497,6 @@ use Carp;
 our @ISA = "PDL::PP::Block";
 our @CARP_NOT;
 
-
 sub new {
 	my $type   = shift;
 	bless [],$type;
@@ -520,42 +508,37 @@ sub myprelude {
     #  Set appropriate function from the vtable to supply to threadthreadloop.
     #    Function name from the vtable is readdata for normal code
     #    function name for backcode is writebackdata
-    my $funcName = "readdata";
-    $funcName = "writebackdata" if( $backcode );
-
+    my $funcName = $backcode ? "writebackdata" : "readdata";
     my ($ord,$pdls) = $parent->get_pdls();
-
-    my $str = '';
     my $loop_key = "emitted_startthreadloop_$funcName";
     my $macro_name = "PDL_STARTTHREADLOOP_$parent->{Name}_$funcName";
-    if (!$parent->{$loop_key}) {
-	$parent->{$loop_key} = 1;
-	$str .= PDL::PP::pp_line_numbers(__LINE__, join " \\\n\t",
-	  "#define $macro_name",
-	  'if ( PDL->startthreadloop(&($PRIV(__pdlthread)),$PRIV(vtable)->'.$funcName.', __tr) ) return; \
-	   do { register PDL_Indx __tind1=0,__tind2=0; \
-		register PDL_Indx __tnpdls = $PRIV(__pdlthread).npdls; \
-		register PDL_Indx __tdims1 = $PRIV(__pdlthread.dims[1]); \
-		register PDL_Indx __tdims0 = $PRIV(__pdlthread.dims[0]); \
-		register PDL_Indx *__offsp = PDL->get_threadoffsp(&$PRIV(__pdlthread));',
-	  ( map { "register PDL_Indx __tinc0_${_} = \$PRIV(__pdlthread).incs[${_}];"} 0..$#{$ord}),
-	  ( map { "register PDL_Indx __tinc1_${_} = \$PRIV(__pdlthread).incs[__tnpdls+$_];"} 0.. $#{$ord}),
-	  ( map { $ord->[$_] ."_datap += __offsp[$_];"} 0..$#{$ord} ),
-	  'for( __tind2 = 0 ; \
-		__tind2 < __tdims1 ; \
-		__tind2++',
-		( map { "\t\t," . $ord->[$_] . "_datap += __tinc1_${_} - __tinc0_${_} * __tdims0"} 0..$#{$ord} ),
-	     ')',
-	  '{ \
-	     for( __tind1 = 0 ; \
-		  __tind1 < __tdims0 ; \
-		  __tind1++',
-		  ( map { "\t\t," . $ord->[$_] . "_datap += __tinc0_${_}"} 0..$#{$ord}),
-	       ") {",
-	  "PDL_COMMENT(\"This is the tightest threadloop. Make sure inside is optimal.\")\n\n",
-	);
-    }
-    $str . $macro_name;
+    return $macro_name if $parent->{$loop_key};
+    $parent->{$loop_key} = 1;
+    PDL::PP::pp_line_numbers(__LINE__, join " \\\n\t",
+      "#define $macro_name",
+      'if ( PDL->startthreadloop(&($PRIV(pdlthread)),$PRIV(vtable)->'.$funcName.', __tr) ) return; \
+       do { \
+	    PDL_Indx *__tdims = PDL->get_threaddims(&$PRIV(pdlthread)); \
+	    register PDL_Indx __tdims1 = __tdims[1]; \
+	    register PDL_Indx __tdims0 = __tdims[0]; \
+	    register PDL_Indx *__offsp = PDL->get_threadoffsp(&$PRIV(pdlthread));',
+      'PDL_COMMENT("incs are each pdl\'s stride, declared at func start")',
+      'PDL_COMMENT("offs are each pthread\'s starting offset into each pdl")',
+      (map $pdls->{$ord->[$_]}->do_pointeraccess." += __offsp[$_];", 0..$#$ord),
+      'for( __tind1 = 0 ; \
+	    __tind1 < __tdims1 ; \
+	    __tind1++',
+	    'PDL_COMMENT("step by tinc1, undoing inner-loop of tinc0*tdims0")',
+	    (map "\t\t,".$pdls->{$ord->[$_]}->do_pointeraccess." += __tinc1_$_ - __tinc0_$_ * __tdims0", 0..$#$ord),
+	 ')',
+      '{ \
+	 for( __tind0 = 0 ; \
+	      __tind0 < __tdims0 ; \
+	      __tind0++',
+	      (map "\t\t,".$pdls->{$ord->[$_]}->do_pointeraccess." += __tinc0_${_}", 0..$#{$ord}),
+	   ") {",
+      "PDL_COMMENT(\"This is the tightest threadloop. Make sure inside is optimal.\")\n\n",
+    ) . $macro_name;
 }
 
 # Should possibly fold out thread.dims[0] and [1].
@@ -563,18 +546,16 @@ sub mypostlude {my($this,$parent,$context) = @_;
     my ($ord,$pdls) = $parent->get_pdls();
     my $loop_key = "emitted_endthreadloop";
     my $macro_name = "PDL_ENDTHREADLOOP_$parent->{Name}";
-    my $str = '';
-    if (!$parent->{$loop_key}) {
-	$parent->{$loop_key} = 1;
-	$str .= PDL::PP::pp_line_numbers(__LINE__, join " \\\n\t",
-	    "\n#define $macro_name",
-	    '}',
-	    '}',
-	    ( map { $ord->[$_] . "_datap -= __tinc1_${_} * __tdims1 + __offsp[${_}];"} 0..$#{$ord} ),
-	    '} while(PDL->iterthreadloop(&$PRIV(__pdlthread),2));'."\n",
-	    );
-    }
-    $str . $macro_name;
+    return $macro_name if $parent->{$loop_key};
+    $parent->{$loop_key} = 1;
+    PDL::PP::pp_line_numbers(__LINE__, join " \\\n\t",
+	"\n#define $macro_name",
+	'}',
+	'}',
+	'PDL_COMMENT("undo outer-loop of tinc1*tdims1, and original per-pthread offset")',
+	(map $pdls->{$ord->[$_]}->do_pointeraccess." -= __tinc1_$_ * __tdims1 + __offsp[$_];", 0..$#$ord),
+	'} while(PDL->iterthreadloop(&$PRIV(pdlthread),2));'."\n",
+	) . $macro_name;
 }
 
 # Simple subclass of ThreadLoop to implement writeback code
@@ -614,7 +595,7 @@ sub new {
 sub myoffs { return 1; }
 sub myprelude {
     my($this,$parent,$context) = @_;
-    return "\n#if ". (join '||',map {"(THISIS_$_(1)+0)"} split '',$this->[0])."\n";
+    return "\n#if ". (join '||',map "(THISIS_$_(1)+0)", split '',$this->[0])."\n";
 }
 
 sub mypostlude {my($this,$parent,$context) = @_;
@@ -641,6 +622,7 @@ sub new { my($type,$str,$parent) = @_;
     elsif($pdl =~ /^SIZE$/) {PDL::PP::SizeAccess->new($pdl,$inds);}
     elsif($pdl =~ /^RESIZE$/) {PDL::PP::ReSizeAccess->new($pdl,$inds);}
     elsif($pdl =~ /^GENERIC$/) {PDL::PP::GentypeAccess->new($pdl,$inds);}
+    elsif($pdl =~ /^PPSYM$/) {PDL::PP::PpsymAccess->new($pdl,$inds);}
     elsif($pdl =~ /^PDL$/) {PDL::PP::PdlAccess->new($pdl,$inds);}
     elsif(!defined $parent->{ParObjs}{$pdl}) {PDL::PP::OtherAccess->new($pdl,$inds);}
     else {
@@ -929,7 +911,7 @@ our @CARP_NOT;
 sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
 
 sub get_str {my($this,$parent,$context) = @_;
-	croak ("can't access undefined pdl ".$this->[0])
+	confess ("can't access undefined pdl ".$this->[0])
 	  unless defined($parent->{ParObjs}{$this->[0]});
 #	$parent->{ParObjs}{$this->[0]}->{FlagPaccess} = 1;
 	$parent->{ParObjs}{$this->[0]}->{FlagPhys} = 1;
@@ -963,7 +945,7 @@ our @CARP_NOT;
 sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
 
 sub get_str {my($this,$parent,$context) = @_;
-	croak ("can't access undefined pdl ".$this->[0])
+	confess ("can't access undefined pdl ".$this->[0])
 	  unless defined($parent->{ParObjs}{$this->[0]});
 	$parent->{ParObjs}{$this->[0]}->do_pdlaccess();
 }
@@ -1014,48 +996,11 @@ our @CARP_NOT;
 sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
 
 sub get_str {my($this,$parent,$context) = @_;
-	croak "can't get SIZE of undefined dimension $this->[0]"
+	confess "can't get SIZE of undefined dimension $this->[0]"
 	  unless defined($parent->{IndObjs}{$this->[0]});
 	$parent->{IndObjs}{$this->[0]}->get_size();
 }
 
-###########################
-#
-# Encapsulate a ReSizeAccess
-
-package PDL::PP::ReSizeAccess;
-use Carp;
-our @CARP_NOT;
-
-sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
-
-sub get_str {my($this,$parent,$context) = @_;
-	$this->[0] =~ /^([^,]+),([^,]+)$/ or
-		croak "Can't interpret resize str $this->[0]";
-	croak "can't RESIZE undefined dimension $1"
-	  unless defined($parent->{IndObjs}{$1});
-
-	my $s = $parent->{IndObjs}{$1}->get_size();
-
-# XXX NOTE: All ndarrays must be output ndarrays, there must not be
-# a loop over this var (at all!) etc. Should check for these,
-# this is why not yet documented.
-# FURTHER NOTE: RESIZE DOESN'T COPY DATA PROPERLY!
-
-	my($ord,$pdls) = $parent->get_pdls();
-	my @p;
-
-	for(@$ord) {
-		push @p, $_
-			if $pdls->{$_}->has_dim($1);
-	}
-	print "RESIZEACC: $1 $2, (",(join ',',@p),")\n";
-	warn "RESIZE USED: DO YOU KNOW WHAT YOU ARE DOING???\n";
-
-	return "$s = $2; ".(join '',map {$pdls->{$_}->do_resize($1,$2)} @p);
-}
-
-
 ###########################
 #
 # Encapsulate a GentypeAccess
@@ -1067,41 +1012,25 @@ our @CARP_NOT;
 sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
 
 sub get_str {my($this,$parent,$context) = @_;
-  croak "generic type access outside a generic loop"
+  confess "generic type access outside a generic loop"
     unless defined(my $type = $parent->{Gencurtype}[-1]);
   return $type->ctype if !$this->[0];
-  my $pobj = $parent->{ParObjs}{$this->[0]} // croak "not a defined parname";
+  my $pobj = $parent->{ParObjs}{$this->[0]} // confess "not a defined parname";
   $pobj->adjusted_type($type)->ctype;
 }
 
-########################
-#
-# Type coercion
-#
-# Now, if TYPES:F given and double arguments, will coerce.
+package PDL::PP::PpsymAccess;
+use Carp;
+our @CARP_NOT;
 
-package PDL::PP::TypeConv;
+sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
 
-sub print_xscoerce { my($this) = @_;
-	$this->printxs("\t__priv->datatype=PDL_B;\n");
-# First, go through all the types, selecting the most general.
-	for(@{$this->{PdlOrder}}) {
-		$this->printxs($this->{Pdls}{$_}->get_xsdatatypetest());
-	}
-# See which types we are allowed to use.
-	$this->printxs("\tif(0) {}\n");
-	for(@{$this->get_generictypes()}) {
-		$this->printxs("\telse if(__priv->datatype <= $_->[2]) __priv->datatype = $_->[2];\n");
-	}
-	$this->{Types} =~ /F/ and (
-		$this->printxs("\telse if(__priv->datatype == PDL_D) {__priv->datatype = PDL_F; PDL_COMMENT(\"Cast double to float\")}\n"));
-	$this->printxs(qq[\telse {croak("Too high type \%d given!\\n",__priv->datatype);}]);
-# Then, coerce everything to this type.
-	for(@{$this->{PdlOrder}}) {
-		$this->printxs($this->{Pdls}{$_}->get_xscoerce());
-	}
+sub get_str {my($this,$parent,$context) = @_;
+  confess "generic type access outside a generic loop"
+    unless defined(my $type = $parent->{Gencurtype}[-1]);
+  return $type->ppsym if !$this->[0];
+  my $pobj = $parent->{ParObjs}{$this->[0]} // confess "not a defined parname";
+  $pobj->adjusted_type($type)->ctype;
 }
-# XXX Should use PDL::Core::Dev;
 
-# return true
 1;
